@@ -3,6 +3,15 @@
 #include <math.h>
 #include "gsl_nls.h"
 
+/* static function declarations */
+static int gsl_f(const gsl_vector *x, void *params, gsl_vector *f);
+
+static int gsl_df(const gsl_vector *x, void *params, gsl_matrix *J);
+
+static int gsl_fvv(const gsl_vector *x, const gsl_vector *v, void *params, gsl_vector *fvv);
+
+static void callback(const R_len_t iter, void *params, const gsl_multifit_nlinear_workspace *w);
+
 /* cleanup memory */
 static void C_nls_cleanup(void *data)
 {
@@ -15,21 +24,38 @@ static void C_nls_cleanup(void *data)
         gsl_qrng_free(pars->q);
     if (pars->wts)
         gsl_vector_free(pars->wts);
+    if (pars->Lw)
+        gsl_matrix_free(pars->Lw);
     if (pars->mx)
         gsl_matrix_free(pars->mx);
-    if (pars->mp)
-        gsl_vector_free(pars->mp);
     if (pars->mpopt)
         gsl_vector_free(pars->mpopt);
     if (pars->diag)
         gsl_vector_free(pars->diag);
+    if (pars->lu)
+        gsl_matrix_free(pars->lu);
+    if (pars->JTJ)
+        gsl_matrix_free(pars->JTJ);
+    if (pars->workn)
+        gsl_vector_free(pars->workn);
+    if (pars->workp)
+        gsl_vector_free(pars->workp);
+    if (pars->worknp)
+        gsl_matrix_free(pars->worknp);
+    if (pars->mpopt1)
+        gsl_vector_free(pars->mpopt1);
+    if (pars->psi)
+        gsl_vector_free(pars->psi);
+    if (pars->psip)
+        gsl_vector_free(pars->psip);
 }
 
 /* function call w/ cleanup */
-SEXP C_nls(SEXP fn, SEXP y, SEXP jac, SEXP fvv, SEXP env, SEXP start, SEXP swts, SEXP control_int, SEXP control_dbl)
+SEXP C_nls(SEXP fn, SEXP y, SEXP jac, SEXP fvv, SEXP env, SEXP start, SEXP swts, SEXP lupars, SEXP control_int, SEXP control_dbl, SEXP has_start, SEXP loss_config)
 {
     /* function arguments */
-    pdata pars = {fn, y, jac, fvv, env, start, swts, control_int, control_dbl, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
+    pdata pars = {fn, y, jac, fvv, env, start, swts, lupars, control_int, control_dbl, has_start, loss_config,
+                  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
 
     /* safe function call */
     SEXP ans = R_ExecWithCleanup(C_nls_internal, &pars, C_nls_cleanup, &pars);
@@ -144,6 +170,8 @@ SEXP C_nls_internal(void *data)
     params.y = pars->y;
     params.rho = pars->env;
     params.start = pars->start;
+    params.startisnum = INTEGER_ELT(pars->control_int, 13);
+
     if (!mstart)
         params.warn = TRUE;
     else
@@ -188,17 +216,59 @@ SEXP C_nls_internal(void *data)
     }
 
     /* use weights */
+    int wgt_i = INTEGER_ELT(VECTOR_ELT(pars->loss_config, 0), 0);
     pars->wts = gsl_vector_alloc(n);
+    gsl_vector_set_all(pars->wts, 1.0);
     if (!Rf_isNull(pars->swts))
-        for (R_len_t i = 0; i < n; i++)
-            gsl_vector_set(pars->wts, i, REAL_ELT(pars->swts, i));
+    {
+        if (Rf_isMatrix(pars->swts))
+        {
+            /* move diagonal to pars->wts */
+            pars->Lw = gsl_matrix_alloc(n, n);
+            double *swts_ptr = REAL(pars->swts);
+            for (R_len_t n1 = 0; n1 < n; n1++)
+            {
+                gsl_vector_set(pars->wts, n1, swts_ptr[n1 + n * n1] * swts_ptr[n1 + n * n1]);
+                for (R_len_t n2 = 0; n2 < n; n2++)
+                    gsl_matrix_set(pars->Lw, n1, n2, swts_ptr[n1 + n * n2] / swts_ptr[n1 + n * n1]);
+            }
+        }
+        else
+        {
+            double *swts_ptr = REAL(pars->swts);
+            for (R_len_t i = 0; i < n; i++)
+                gsl_vector_set(pars->wts, i, swts_ptr[i] * swts_ptr[i]);
+        }
+    }
 
     /* initialize solver */
     const gsl_multifit_nlinear_type *T = gsl_multifit_nlinear_trust;
 
+    /* parameter constraints */
+    if (Rf_isMatrix(pars->lupars))
+    {
+        double *luptr = REAL(pars->lupars);
+        pars->lu = gsl_matrix_alloc(2, p);
+        for (R_len_t k = 0; k < p; k++)
+        {
+            if (R_finite(luptr[2 * k]))
+                gsl_matrix_set(pars->lu, 0, k, luptr[2 * k]);
+            else
+                gsl_matrix_set(pars->lu, 0, k, (double)GSL_NEGINF);
+            if (R_finite(luptr[2 * k + 1]))
+                gsl_matrix_set(pars->lu, 1, k, luptr[2 * k + 1]);
+            else
+                gsl_matrix_set(pars->lu, 1, k, (double)GSL_POSINF);
+        }
+    }
+
     /* allocate workspace with default parameters */
     pars->w = gsl_multifit_nlinear_alloc(T, &fdf_params, n, p);
     pars->mpopt = gsl_vector_alloc(p);
+    pars->workp = gsl_vector_alloc(p);
+    pars->workn = gsl_vector_alloc(n);
+    pars->worknp = gsl_matrix_alloc(n, p);
+    gsl_vector_set_zero(pars->mpopt);
 
     /* multistart algorithm */
     if (mstart)
@@ -209,6 +279,20 @@ SEXP C_nls_internal(void *data)
         else
             pars->q = gsl_qrng_alloc(gsl_qrng_halton, p);
 
+        /* manually initialize workspace */
+        (pars->w)->fdf = &fdf;
+        (pars->w)->sqrt_wts = (pars->w)->sqrt_wts_work;
+        if (!Rf_isNull(pars->swts))
+        {
+            for (R_len_t i = 0; i < n; i++)
+            {
+                double wi = gsl_vector_get(pars->wts, i);
+                gsl_vector_set((pars->w)->sqrt_wts, i, sqrt(wi));
+            }
+        }
+        else
+            gsl_vector_set_all((pars->w)->sqrt_wts, 1.0);
+
         /* multistart parameters */
         mdata mpars;
         mpars.n = INTEGER_ELT(pars->control_int, 6);
@@ -218,17 +302,34 @@ SEXP C_nls_internal(void *data)
         mpars.niter = INTEGER_ELT(pars->control_int, 10);
         mpars.max = INTEGER_ELT(pars->control_int, 11);
         mpars.minsp = INTEGER_ELT(pars->control_int, 12);
+        mpars.wgt_i = wgt_i;
+        mpars.all_start = TRUE;
+        mpars.has_start = LOGICAL(pars->has_start);
         mpars.r = REAL_ELT(pars->control_dbl, 8);
         mpars.tol = REAL_ELT(pars->control_dbl, 9);
+        mpars.dtol = 1.0e-6;
         mpars.ntix = (int *)S_alloc(mpars.n, sizeof(int));
         mpars.qmp = (double *)S_alloc(p, sizeof(double));
-        mpars.startptr = REAL(pars->start);
         mpars.mssr_order = (int *)R_alloc(mpars.n, sizeof(int));
         mpars.mstop = GSL_CONTINUE;
         mpars.mstarts = 0;
         mpars.nsp = 0;
         mpars.nwsp = 0;
-        mpars.mssropt = (double)GSL_POSINF;
+        mpars.luchange = (int *)S_alloc(p, sizeof(int));
+        mpars.rejectscl = 1.25;
+        mpars.mssropt[0] = (double)GSL_POSINF;
+        mpars.mssropt[1] = (double)GSL_POSINF;
+        mpars.ssrconv[0] = 1.0;
+        mpars.ssrconv[1] = 1.0;
+        mpars.start = (double *)R_alloc(2 * p, sizeof(double));
+        mpars.maxlims = (double *)R_alloc(2 * p, sizeof(double));
+
+        // mpoptinfo mpoptall;
+        // mpoptall.mpcount = 0;
+        // mpoptall.mpmax = 100;
+        // mpoptall.mpall = (double *)S_alloc(mpars.p * mpoptall.mpmax, sizeof(double));
+        // mpoptall.mpradii = (double *)S_alloc(mpars.p * mpoptall.mpmax, sizeof(double));
+        // mpars.mpopt = &mpoptall;
 
         SEXP mssr = PROTECT(Rf_allocVector(REALSXP, mpars.n));
         SEXP mpar = PROTECT(Rf_allocVector(REALSXP, p));
@@ -236,27 +337,176 @@ SEXP C_nls_internal(void *data)
         params.start = mpar;
         nprotect += 2;
 
-        /* placeholder matrices/vectors */
+        /* workspace matrices/vectors */
         pars->mx = gsl_matrix_alloc(mpars.n, p);
-        pars->mp = gsl_vector_alloc(p);
         pars->diag = gsl_vector_alloc(p);
+        pars->JTJ = gsl_matrix_alloc(p, p);
+        pars->mpopt1 = gsl_vector_alloc(p);
+
+        double *startptr = REAL(pars->start);
+        for (R_len_t k = 0; k < p; k++)
+        {
+            mpars.start[2 * k] = startptr[2 * k];
+            mpars.start[2 * k + 1] = startptr[2 * k + 1];
+            mpars.maxlims[2 * k] = startptr[2 * k];
+            mpars.maxlims[2 * k + 1] = startptr[2 * k + 1];
+        }
+
         /* initially focus sampling around center points */
         for (R_len_t k = 0; k < p; k++)
-            gsl_vector_set(pars->diag, k, (mpars.startptr[2 * k + 1] - mpars.startptr[2 * k]) / 20.0);
+        {
+            if (!(mpars.has_start)[2 * k] || !(mpars.has_start)[2 * k + 1])
+            {
+                gsl_vector_set(pars->diag, k, 1.0);
+                mpars.all_start = FALSE;
+            }
+            else
+            {
+                gsl_vector_set(pars->diag, k, 0.75);
+                if ((mpars.start)[2 * k] + xtol > (mpars.start)[2 * k + 1])
+                    mpars.rejectscl = -1.0; // starting range is not an interval
+            }
+        }
 
         /* multi-start global iterations */
         do
         {
-            gsl_multistart_driver(pars, &mpars, &fdf, mssr, xtol, ftol, gtol, verbose);
+            gsl_multistart_driver(pars, &mpars, &fdf, mssr, xtol, ftol, FALSE, verbose);
 
             /* check stopping criterion */
             mpars.mstarts += 1;
+
             if (mpars.mstarts > mpars.max)
                 mpars.mstop = GSL_EMAXITER;
-            if (mpars.nsp >= mpars.minsp && mpars.nwsp >= mpars.r * mpars.nsp)
+
+            if (mpars.nsp >= mpars.minsp && mpars.nwsp > (mpars.r + sqrt(mpars.r) * mpars.nsp))
                 mpars.mstop = GSL_SUCCESS;
 
+            if (!(mpars.mstarts % 10) && !(mpars.mssropt[0] < (double)GSL_POSINF))
+            {
+                /* reduce determinant tolerance */
+                mpars.dtol = gsl_max(0.5 * mpars.dtol, GSL_DBL_EPSILON);
+                if (!(mpars.mstarts % 100)) // reset start ranges
+                {
+                    for (R_len_t k = 0; k < p; k++)
+                    {
+                        mpars.start[2 * k] = startptr[2 * k];
+                        mpars.start[2 * k + 1] = startptr[2 * k + 1];
+                    }
+                }
+            }
+
         } while (mpars.mstop == GSL_CONTINUE);
+        /* second pass multi-start global iterations */
+        if (wgt_i)
+        {
+            /*
+                outliers are identified as observations with Di > min(4 / n, 5 * mad(Di)),
+                associated weights are set to zero in the second pass of the multi-start procedure.
+            */
+            if (mpars.mssropt[1] < mpars.mssropt[0])
+                gsl_vector_memcpy(pars->mpopt, pars->mpopt1);
+            // compute J and J^T * J using pars->mpopt
+            if (pars->Lw)
+                gsl_multifit_nlinear_winit_LD(pars->mpopt, pars->wts, pars->Lw, &fdf, pars->w);
+            else
+                gsl_multifit_nlinear_winit(pars->mpopt, pars->wts, &fdf, pars->w);
+
+            det_eval_jtj((pars->w)->params, (pars->w)->sqrt_wts, pars->Lw, (pars->w)->fdf,
+                         (pars->w)->x, (pars->w)->f, (pars->w)->J, pars->JTJ, pars->workn);
+
+            // compute cook's d
+            mpars.mstop = cooks_d((pars->w)->f, (pars->w)->J, pars->JTJ, pars->workn, pars->worknp);
+
+            if (!mpars.mstop)
+            {
+                // update weights
+                R_len_t noutlier = 0;
+                double *workn = (double *)R_alloc(n, sizeof(double));
+                double mad = gsl_stats_mad((pars->workn)->data, 1, n, workn);
+                double thresh = gsl_min(4.0 / n, 5 * mad);
+                (pars->w)->sqrt_wts = (pars->w)->sqrt_wts_work;
+                for (R_len_t i = 0; i < n; i++)
+                {
+                    double di = gsl_vector_get(pars->workn, i);
+                    if (di > thresh)
+                    {
+                        gsl_vector_set(pars->wts, i, 0.0);
+                        gsl_vector_set((pars->w)->sqrt_wts, i, 0.0);
+                        noutlier += 1;
+                    }
+                    else
+                    {
+                        double wi = gsl_vector_get(pars->wts, i);
+                        gsl_vector_set((pars->w)->sqrt_wts, i, sqrt(wi));
+                    }
+                }
+                if ((noutlier > 0) && (noutlier < (n - p)))
+                {
+                    /* reset qrng workspace */
+                    gsl_qrng_init(pars->q);
+                    // reset counters
+                    mpars.mstop = GSL_CONTINUE;
+                    mpars.mstarts = 0;
+                    mpars.nsp = 0;
+                    mpars.nwsp = 0;
+                    mpars.dtol = 1.0e-6;
+                    mpars.rejectscl = 1.25;
+                    mpars.mssropt[0] = (double)GSL_POSINF;
+                    mpars.mssropt[1] = (double)GSL_POSINF;
+                    mpars.ssrconv[0] = 1.0;
+                    mpars.ssrconv[1] = 1.0;
+                    memset(mpars.ntix, 0, mpars.n * sizeof(int));
+                    memset(mpars.luchange, 0, p * sizeof(int));
+                    // second pass multi-start
+                    do
+                    {
+                        gsl_multistart_driver(pars, &mpars, &fdf, mssr, xtol, ftol, TRUE, verbose);
+
+                        /* check stopping criterion */
+                        mpars.mstarts += 1;
+
+                        if (mpars.mstarts > mpars.max)
+                            mpars.mstop = GSL_EMAXITER;
+
+                        if (mpars.nsp >= mpars.minsp && mpars.nwsp > (mpars.r + sqrt(mpars.r) * mpars.nsp))
+                            mpars.mstop = GSL_SUCCESS;
+
+                        if (!(mpars.mstarts % 10) && !(mpars.mssropt[0] < (double)GSL_POSINF))
+                        {
+                            /* reduce determinant tolerance */
+                            mpars.dtol = gsl_max(0.5 * mpars.dtol, GSL_DBL_EPSILON);
+                            if (!(mpars.mstarts % 100)) // reset start ranges
+                            {
+                                for (R_len_t k = 0; k < p; k++)
+                                {
+                                    mpars.start[2 * k] = startptr[2 * k];
+                                    mpars.start[2 * k + 1] = startptr[2 * k + 1];
+                                }
+                            }
+                        }
+                    } while (mpars.mstop == GSL_CONTINUE);
+                }
+                // reset original weights
+                // sqrt_wts is initialized by gsl_multifit_nlinear_winit
+                gsl_vector_set_all(pars->wts, 1.0);
+                if (!Rf_isNull(pars->swts))
+                {
+                    if (Rf_isMatrix(pars->swts))
+                    {
+                        double *swts_ptr = REAL(pars->swts);
+                        for (R_len_t i = 0; i < n; i++)
+                            gsl_vector_set(pars->wts, i, swts_ptr[i + n * i] * swts_ptr[i + n * i]);
+                    }
+                    else
+                    {
+                        double *swts_ptr = REAL(pars->swts);
+                        for (R_len_t i = 0; i < n; i++)
+                            gsl_vector_set(pars->wts, i, swts_ptr[i] * swts_ptr[i]);
+                    }
+                }
+            }
+        }
         if (verbose)
         {
             if (mpars.mstop == GSL_SUCCESS)
@@ -265,10 +515,19 @@ SEXP C_nls_internal(void *data)
                 Rprintf("multi-start algorithm reached max. number of global iterations (NSP = %d, NWSP = %d, # iterations = %d)\n", mpars.nsp, mpars.nwsp, mpars.mstarts);
             Rprintf("*******************\n");
         }
-        /* add noise in case of zero residuals */
-        if(mpars.mssropt <= 0)
+        if (mpars.mssropt[1] < mpars.mssropt[0])
         {
-            gsl_vector_set(pars->mpopt, 0, gsl_vector_get(pars->mpopt, 0) + xtol);
+            mpars.mssropt[0] = mpars.mssropt[1];
+            mpars.ssrconv[0] = mpars.ssrconv[1];
+            gsl_vector_memcpy(pars->mpopt, pars->mpopt1);
+        }
+        /* add small jitter to parameters */
+        if (mpars.mssropt[0] < ftol || mpars.ssrconv[0] < ftol)
+        {
+            if (pars->lu)
+                gsl_vector_set(pars->mpopt, 0, gsl_min(gsl_vector_get(pars->mpopt, 0) + 1.0e-4, gsl_matrix_get(pars->lu, 1, 0)));
+            else
+                gsl_vector_set(pars->mpopt, 0, gsl_vector_get(pars->mpopt, 0) + 1.0e-4);
         }
     }
     else /* single-start */
@@ -278,7 +537,9 @@ SEXP C_nls_internal(void *data)
     }
 
     /* (re-)initialize solver w/ optimal start parameters */
-    if (!Rf_isNull(pars->swts))
+    if (pars->Lw)
+        gsl_multifit_nlinear_winit_LD(pars->mpopt, pars->wts, pars->Lw, &fdf, pars->w);
+    else if (!Rf_isNull(pars->swts) || wgt_i)
         gsl_multifit_nlinear_winit(pars->mpopt, pars->wts, &fdf, pars->w);
     else
         gsl_multifit_nlinear_init(pars->mpopt, &fdf, pars->w);
@@ -301,7 +562,39 @@ SEXP C_nls_internal(void *data)
 
     /* solve the system  */
     int info = GSL_CONTINUE;
-    int status = gsl_multifit_nlinear_driver2(niter, xtol, gtol, ftol, verbose ? callback : NULL, verbose ? &params : NULL, &info, &chisq0, &chisq1, pars->w);
+    int status = GSL_FAILURE;
+    int irls_status = GSL_FAILURE;
+    R_len_t irls_iter = 0;
+    double irls_delta = 0.0;
+    double irls_sigma = 1.0;
+
+    if (!wgt_i) /* default loss function */
+    {
+        status = gsl_multifit_nlinear_driver2(niter, xtol, gtol, ftol, verbose ? callback : NULL,
+                                              verbose ? &params : NULL, &info, &chisq0, &chisq1,
+                                              pars->lu, pars->Lw, pars->w);
+    }
+    else /* non-default loss function */
+    {
+        /* irls psi and psi' */
+        pars->psi = gsl_vector_alloc(n);
+        pars->psip = gsl_vector_alloc(n);
+
+        status = gsl_multifit_nlinear_rho_driver(pars, &fdf, wgt_i, niter, xtol, gtol, ftol, &params,
+                                                 &info, &chisq0, &chisq1, &irls_sigma, &irls_iter, &irls_status, verbose);
+
+        /* copy irls weights */
+        gsl_vector_memcpy(pars->wts, pars->workn);
+
+        /* irls x tolerance */
+        for (R_len_t k = 0; k < p; k++)
+        {
+            double x0 = gsl_vector_get(pars->workp, k);
+            double x1 = gsl_vector_get((pars->w)->x, k);
+            irls_delta = gsl_max(irls_delta, fabs(x0 - x1));
+        }
+    }
+
     R_len_t iter = (R_len_t)gsl_multifit_nlinear_niter(pars->w);
 
     /* compute covariance and cost at best fit parameters */
@@ -318,16 +611,22 @@ SEXP C_nls_internal(void *data)
     {
         /* print summary statistics*/
         Rprintf("*******************\nsummary from method 'multifit/%s'\n", gsl_multifit_nlinear_trs_name(pars->w));
+        if (wgt_i)
+        {
+            Rprintf("IRLS number of iterations: %d\n", irls_iter);
+            Rprintf("IRLS achieved tolerance: %g\n", irls_delta);
+            Rprintf("IRLS convergence status: %s\n", gsl_strerror(irls_status));
+        }
         Rprintf("number of iterations: %d\n", iter);
-        Rprintf("reason for stopping: %s\n", gsl_strerror(info));
-        Rprintf("initial ssr = %g\n", chisq_init);
-        Rprintf("final ssr = %g\n", chisq1);
-        Rprintf("ssr/dof = %g\n", chisq1 / (n - p));
-        Rprintf("ssr achieved tolerance = %g\n", chisq0 - chisq1);
+        // Rprintf("reason for stopping: %s\n", gsl_strerror(info));
+        Rprintf("initial ssr: %g\n", chisq_init);
+        Rprintf("final ssr: %g\n", chisq1);
+        Rprintf("ssr/dof: %g\n", chisq1 / (n - p));
+        Rprintf("ssr achieved tolerance: %g\n", chisq0 - chisq1);
         Rprintf("function evaluations: %d\n", (int)fdf.nevalf);
         Rprintf("jacobian evaluations: %d\n", (int)fdf.nevaldf);
         Rprintf("fvv evaluations: %d\n", (int)fdf.nevalfvv);
-        Rprintf("status = %s\n*******************\n", gsl_strerror(status));
+        Rprintf("status: %s\n*******************\n", gsl_strerror(status));
     }
 
     /* initialize result */
@@ -335,13 +634,13 @@ SEXP C_nls_internal(void *data)
     if (verbose)
     {
         const char *ansnms[] = {"par", "covar", "resid", "grad", "niter", "status", "conv", "ssr", "ssrtol",
-                                "algorithm", "neval", "partrace", "ssrtrace", ""};
+                                "algorithm", "neval", "irls", "partrace", "ssrtrace", ""};
         ans = PROTECT(Rf_mkNamed(VECSXP, ansnms));
     }
     else
     {
         const char *ansnms[] = {"par", "covar", "resid", "grad", "niter", "status", "conv", "ssr", "ssrtol",
-                                "algorithm", "neval", ""};
+                                "algorithm", "neval", "irls", ""};
         ans = PROTECT(Rf_mkNamed(VECSXP, ansnms));
     }
     nprotect++;
@@ -453,6 +752,44 @@ SEXP C_nls_internal(void *data)
     SET_VECTOR_ELT(ans, 10, ansneval);
     UNPROTECT(1);
 
+    /* irls weights */
+    if (wgt_i)
+    {
+        const char *irlsnms[] = {"irls_weights", "irls_psi", "irls_dpsi", "irls_sigma", "irls_status", "irls_niter", "irls_tol", "irls_conv", ""};
+        SEXP ansirls = PROTECT(Rf_mkNamed(VECSXP, irlsnms));
+        SEXP irlswts = PROTECT(Rf_allocVector(REALSXP, n));
+        SEXP irlspsi = PROTECT(Rf_allocVector(REALSXP, n));
+        SEXP irlspsip = PROTECT(Rf_allocVector(REALSXP, n));
+        if (status == GSL_SUCCESS || status == GSL_EMAXITER)
+        {
+            for (R_len_t i = 0; i < n; i++)
+            {
+                SET_REAL_ELT(irlswts, i, gsl_vector_get(pars->wts, i));
+                SET_REAL_ELT(irlspsi, i, gsl_vector_get(pars->psi, i));
+                SET_REAL_ELT(irlspsip, i, gsl_vector_get(pars->psip, i));
+            }
+        }
+        else
+        {
+            for (R_len_t i = 0; i < n; i++)
+            {
+                SET_REAL_ELT(irlswts, i, NA_REAL);
+                SET_REAL_ELT(irlspsi, i, NA_REAL);
+                SET_REAL_ELT(irlspsip, i, NA_REAL);
+            }
+        }
+        SET_VECTOR_ELT(ansirls, 0, irlswts);
+        SET_VECTOR_ELT(ansirls, 1, irlspsi);
+        SET_VECTOR_ELT(ansirls, 2, irlspsip);
+        SET_VECTOR_ELT(ansirls, 3, Rf_ScalarReal(irls_sigma));
+        SET_VECTOR_ELT(ansirls, 4, Rf_ScalarString(Rf_mkChar(gsl_strerror(irls_status))));
+        SET_VECTOR_ELT(ansirls, 5, Rf_ScalarInteger(irls_iter));
+        SET_VECTOR_ELT(ansirls, 6, Rf_ScalarReal(irls_delta));
+        SET_VECTOR_ELT(ansirls, 7, Rf_ScalarInteger(irls_status));
+        SET_VECTOR_ELT(ans, 11, ansirls);
+        UNPROTECT(4);
+    }
+
     if (verbose)
     {
         if (!Rf_isNull(parnames))
@@ -463,8 +800,8 @@ SEXP C_nls_internal(void *data)
             Rf_setAttrib(params.partrace, R_DimNamesSymbol, dimnames);
             UNPROTECT(1);
         }
-        SET_VECTOR_ELT(ans, 11, params.partrace);
-        SET_VECTOR_ELT(ans, 12, params.ssrtrace);
+        SET_VECTOR_ELT(ans, 12, params.partrace);
+        SET_VECTOR_ELT(ans, 13, params.ssrtrace);
     }
 
     /* free memory */
@@ -475,116 +812,13 @@ SEXP C_nls_internal(void *data)
     return ans;
 }
 
-/*
-gsl_multifit_nlinear_driver2()
-  Iterate the nonlinear least squares solver until completion
-
-Inputs: maxiter  - maximum iterations to allow
-        xtol     - tolerance in step x
-        gtol     - tolerance in gradient
-        ftol     - tolerance in ||f||
-        callback - callback function to call each iteration
-        callback_params - parameters to pass to callback function
-        chisq0   - ssr previous iteration
-        chisq1   - ssr current iteration
-        info     - (output) info flag on why iteration terminated
-                   1 = stopped due to small step size ||dx|
-                   2 = stopped due to small gradient
-                   3 = stopped due to small change in f
-                   GSL_ETOLX = ||dx|| has converged to within machine
-                               precision (and xtol is too small)
-                   GSL_ETOLG = ||g||_inf is smaller than machine
-                               precision (gtol is too small)
-                   GSL_ETOLF = change in ||f|| is smaller than machine
-                               precision (ftol is too small)
-        w        - workspace
-
-Return:
-GSL_SUCCESS if converged
-GSL_EBADFUNC if function evaluation failed
-GSL_MAXITER if maxiter exceeded without converging
-GSL_ENOPROG if no accepted step found on first iteration
-*/
-int gsl_multifit_nlinear_driver2(const R_len_t maxiter,
-                                 const double xtol,
-                                 const double gtol,
-                                 const double ftol,
-                                 void (*callback)(const R_len_t iter, void *params,
-                                                  const gsl_multifit_nlinear_workspace *w),
-                                 void *callback_params,
-                                 int *info,
-                                 double *chisq0,
-                                 double *chisq1,
-                                 gsl_multifit_nlinear_workspace *w)
-{
-    int status = GSL_CONTINUE;
-    R_len_t iter = 0;
-    gsl_vector *f = NULL;
-
-    do
-    {
-        /* current ssr */
-        chisq0[0] = chisq1[0];
-
-        status = gsl_multifit_nlinear_iterate(w);
-
-        /* new ssr */
-        f = gsl_multifit_nlinear_residual(w);
-        gsl_blas_ddot(f, f, chisq1);
-
-        if (callback)
-            ((fdata *)callback_params)->chisq = chisq1[0];
-
-        /*
-         * If the solver reports no progress on the first iteration,
-         * then it didn't find a single step to reduce the
-         * cost function and more iterations won't help so return.
-         *
-         * If we get a no progress flag on subsequent iterations,
-         * it means we did find a good step in a previous iteration,
-         * so continue iterating since the solver has now reset
-         * mu to its initial value.
-         */
-        if (status == GSL_EBADFUNC || (status == GSL_ENOPROG && iter == 0))
-        {
-            *info = status;
-            return status;
-        }
-
-        ++iter;
-
-        if (callback)
-            callback(iter, callback_params, w);
-
-        /* test for convergence */
-        status = gsl_multifit_nlinear_test(xtol, gtol, ftol, info, w);
-    } while (status == GSL_CONTINUE && iter < maxiter);
-
-    /*
-     * the following error codes mean that the solution has converged
-     * to within machine precision, so record the error code in info
-     * and return success
-     */
-    if (status == GSL_ETOLF || status == GSL_ETOLX || status == GSL_ETOLG)
-    {
-        *info = status;
-        status = GSL_SUCCESS;
-    }
-
-    /* check if max iterations reached */
-    if (iter >= maxiter && status != GSL_SUCCESS)
-        status = GSL_EMAXITER;
-
-    return status;
-} /* gsl_multifit_nlinear_driver() */
-
-int gsl_f(const gsl_vector *x, void *params, gsl_vector *f)
+static int gsl_f(const gsl_vector *x, void *params, gsl_vector *f)
 {
     /* construct parameter vector */
     SEXP par = NULL;
     R_len_t p = ((fdata *)params)->p;
     SEXP start = ((fdata *)params)->start;
-    if (Rf_isNumeric(start))
+    if (((fdata *)params)->startisnum)
     {
         par = PROTECT(Rf_allocVector(REALSXP, p));
         for (R_len_t k = 0; k < p; k++)
@@ -627,13 +861,13 @@ int gsl_f(const gsl_vector *x, void *params, gsl_vector *f)
     return GSL_SUCCESS;
 }
 
-int gsl_df(const gsl_vector *x, void *params, gsl_matrix *J)
+static int gsl_df(const gsl_vector *x, void *params, gsl_matrix *J)
 {
     /* construct parameter vector */
     SEXP par = NULL;
     R_len_t p = ((fdata *)params)->p;
     SEXP start = ((fdata *)params)->start;
-    if (Rf_isNumeric(start))
+    if (((fdata *)params)->startisnum)
     {
         par = PROTECT(Rf_allocVector(REALSXP, p));
         for (R_len_t k = 0; k < p; k++)
@@ -681,14 +915,14 @@ int gsl_df(const gsl_vector *x, void *params, gsl_matrix *J)
     return GSL_SUCCESS;
 }
 
-int gsl_fvv(const gsl_vector *x, const gsl_vector *v, void *params, gsl_vector *fvv)
+static int gsl_fvv(const gsl_vector *x, const gsl_vector *v, void *params, gsl_vector *fvv)
 {
     /* populate parameter vector */
     SEXP par = NULL;
     R_len_t p = ((fdata *)params)->p;
     SEXP start = ((fdata *)params)->start;
     SEXP parnames = PROTECT(Rf_getAttrib(start, R_NamesSymbol));
-    if (Rf_isNumeric(start))
+    if (((fdata *)params)->startisnum)
     {
         par = PROTECT(Rf_allocVector(REALSXP, p));
         for (R_len_t k = 0; k < p; k++)
@@ -743,7 +977,7 @@ int gsl_fvv(const gsl_vector *x, const gsl_vector *v, void *params, gsl_vector *
     return GSL_SUCCESS;
 }
 
-void callback(const R_len_t iter, void *params, const gsl_multifit_nlinear_workspace *w)
+static void callback(const R_len_t iter, void *params, const gsl_multifit_nlinear_workspace *w)
 {
     /* update traces */
     double chisq = ((fdata *)params)->chisq;
@@ -760,106 +994,6 @@ void callback(const R_len_t iter, void *params, const gsl_multifit_nlinear_works
         Rprintf((k < (p - 1)) ? "%g, " : "%g)\n", parptr[iter + n * k]);
 }
 
-void gsl_multistart_driver(pdata *pars,
-                           mdata *mpars,
-                           gsl_multifit_nlinear_fdf *fdff,
-                           SEXP mssr,
-                           const double xtol,
-                           const double ftol,
-                           const double gtol,
-                           Rboolean verbose)
-{
-    // initialize variables
-    int minfo;
-    R_len_t p = (R_len_t)(pars->mp->size);
-    double kd, l0, l1, kd0, kd1;
-    double rcond, mchisq0 = (double)GSL_POSINF, mchisq1 = (double)GSL_POSINF;
-    trust_state_t *trust_state = (trust_state_t *)(pars->w->state);
-
-    /* sample initial points */
-    for (R_len_t nn = 0; nn < mpars->n; nn++)
-    {
-        if ((mpars->ntix)[nn] == 0)
-        {
-            gsl_qrng_get(pars->q, mpars->qmp);
-            for (R_len_t k = 0; k < p; k++)
-            {
-                l0 = (mpars->startptr)[2 * k];
-                l1 = (mpars->startptr)[2 * k + 1];
-
-                if (l1 > l0)
-                {
-                    kd = gsl_vector_get(pars->diag, k);
-                    kd0 = 1.0 / (1.0 + exp(kd / 2.0));
-                    kd1 = 1.0 / (1.0 + exp(-kd / 2.0));
-                    // gsl_matrix_set(pars->mx, nn, k, l0 + (l1 - l0) * (mpars->qmp)[k]);
-                    (mpars->qmp)[k] = kd0 + (kd1 - kd0) * (mpars->qmp)[k];
-                    gsl_matrix_set(pars->mx, nn, k, (l1 + l0) / 2.0 + (l1 - l0) / kd * log((mpars->qmp)[k] / (1 - (mpars->qmp)[k])));
-                }
-                else
-                {
-                    gsl_matrix_set(pars->mx, nn, k, l0);
-                }
-            }
-        }
-
-        /* concentrate points */
-        gsl_matrix_get_row(pars->mp, pars->mx, nn);
-        if (!Rf_isNull(pars->swts))
-            gsl_multifit_nlinear_winit(pars->mp, pars->wts, fdff, pars->w);
-        else
-            gsl_multifit_nlinear_init(pars->mp, fdff, pars->w);
-        gsl_multifit_nlinear_driver2(mpars->p, xtol, 1e-3, ftol, NULL, NULL, &minfo, &mchisq0, &mchisq1, pars->w);
-        gsl_matrix_set_row(pars->mx, nn, (pars->w)->x);
-        if (mchisq1 < (double)GSL_POSINF)
-            SET_REAL_ELT(mssr, nn, mchisq1);
-        else
-            SET_REAL_ELT(mssr, nn, NA_REAL);
-    }
-    /* reduce sample points */
-    R_orderVector1(mpars->mssr_order, mpars->n, mssr, TRUE, FALSE);
-    for (R_len_t nn = 0; nn < mpars->n; nn++)
-        (mpars->ntix)[(mpars->mssr_order)[nn]] = (nn < (mpars->q)) ? (mpars->ntix)[(mpars->mssr_order)[nn]] + 1 : 0;
-
-    /* local optimization stage */
-    for (R_len_t nn = 0; nn < (mpars->n); nn++)
-    {
-        if ((mpars->ntix)[nn] >= (mpars->s) && !R_IsNA(REAL_ELT(mssr, nn)))
-        {
-            (mpars->ntix)[nn] = 0;
-            if ((mpars->nsp) == 0 || REAL_ELT(mssr, nn) < (1 + (mpars->tol)) * (mpars->mssropt))
-            {
-                gsl_matrix_get_row(pars->mp, pars->mx, nn);
-                if (!Rf_isNull(pars->swts))
-                    gsl_multifit_nlinear_winit(pars->mp, pars->wts, fdff, pars->w);
-                else
-                    gsl_multifit_nlinear_init(pars->mp, fdff, pars->w);
-                mchisq1 = REAL_ELT(mssr, nn);
-                gsl_multifit_nlinear_driver2(mpars->niter, xtol, 1e-3, ftol, NULL, NULL, &minfo, &mchisq0, &mchisq1, pars->w);
-                gsl_multifit_nlinear_rcond(&rcond, pars->w);
-                if (mchisq1 < (mpars->mssropt) && ((!gsl_isnan(rcond) && rcond > gtol) || mchisq1 < 2 * ftol))
-                {
-                    mpars->mssropt = mchisq1;
-                    mpars->nsp += 1;
-                    mpars->nwsp = 0;
-                    gsl_vector_memcpy(pars->mpopt, (pars->w)->x);
-                    gsl_vector_memcpy(pars->diag, trust_state->diag);
-                    if (verbose)
-                    {
-                        Rprintf("mstart ssr* = %g, cond(J) = %g, NSP = %d, NWSP = %d, par = (", mpars->mssropt, 1.0 / rcond, mpars->nsp, mpars->nwsp);
-                        for (R_len_t k = 0; k < p; k++)
-                            Rprintf((k < (p - 1)) ? "%g, " : "%g)\n", gsl_vector_get((pars->w)->x, k));
-                    }
-                }
-                else
-                    mpars->nwsp += 1;
-            }
-            else
-                mpars->nwsp += 1;
-        }
-    }
-}
-
 /* deprecated function */
 // int gsl_evalf(SEXP par, fdata *params, gsl_vector *f)
 // {
